class MissingParam(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class MissingVar(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class ConfigError(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)



import hashlib

import yaml

from octoploy.config.Config import ProjectConfig, AppConfig, RunMode
from octoploy.oc.Oc import K8Api
from octoploy.utils.Log import Log


class OcObjectDeployer(Log):
    """
    Deploys single openshift objects
    """

    HASH_ANNOTATION = 'yml-hash'

    def __init__(self, root_config: ProjectConfig, oc: K8Api, app_config: AppConfig, mode: RunMode = RunMode()):
        super().__init__()
        self._root_config = root_config  # type: ProjectConfig
        self._app_config = app_config  # type: AppConfig
        self._oc = oc  # type: K8Api
        self._mode = mode

    def select_project(self):
        """
        Selects the required openshift project
        """
        context = self._root_config.get_oc_context()
        if context is not None:
            self._oc.switch_context(context)
        self._oc.project(self._root_config.get_oc_project_name())

    def deploy_object(self, data: dict):
        """
        Deploy the given object (if a deployment required, otherwise does nothing)
        :param data: Data which should be deployed
        """

        # Sort the content so it's always reproducible
        str_repr = yaml.dump(data, sort_keys=True)

        hash_val = hashlib.md5(str_repr.encode('utf-8')).hexdigest()
        metadata = data['metadata']

        # An object might be in a different namespace than the current context
        namespace = metadata.get('namespace')
        if namespace is not None:
            self._oc.project(namespace)

        item_name = data['kind'] + '/' + metadata['name']
        description = self._oc.get(item_name)
        current_hash = None
        if description is not None:
            current_hash = description.get_annotation(self.HASH_ANNOTATION)

        if description is not None and current_hash is None:
            # Item has not been deployed yet with this script, assume both are the same
            self.log.info('Updating annotation of ' + item_name)
            self._oc.annotate(item_name, self.HASH_ANNOTATION, hash_val)
            return

        if current_hash == hash_val:
            self.log.debug('No change in ' + item_name)
            return

        if self._mode.plan:
            self.log.warning('Update required for ' + item_name)
            return

        self.log.info('Applying update ' + item_name + ' (item has changed)')
        self._oc.apply(str_repr)
        self._oc.annotate(item_name, 'yml-hash', hash_val)

        if namespace is not None:
            # Use project namespace as default again
            self._oc.project(self._root_config.get_oc_project_name())

        item_kind = data['kind'].lower()
        if item_kind == 'ConfigMap'.lower():
            self._reload_config()

    def _reload_config(self):
        """
        Tries to reload the configuration for the app
        """
        reload_actions = self._app_config.get_reload_actions()
        for action in reload_actions:
            action.run(self._oc)



import yaml


# define a custom representer for strings
def quoted_presenter(dumper, data):
    return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='"')


class YmlWriter:
    _INIT = False

    @classmethod
    def init(cls):
        if not cls._INIT:
            yaml.add_representer(str, quoted_presenter)
            cls._INIT = True

    @classmethod
    def dump(cls, data, default_flow_style: bool = True) -> str:
        cls.init()
        return yaml.dump(data, sort_keys=True, default_flow_style=default_flow_style)

    @classmethod
    def dump_all(cls, data, file):
        cls.init()
        yaml.dump_all(data, file, default_flow_style=False, sort_keys=True)



from typing import Optional, Dict


class DictUtils:
    @staticmethod
    def get(data: Optional[dict], path: str):
        if data is None:
            return None

        for part in path.split('.'):
            if part not in data:
                return None
            data = data[part]
        return data

    @staticmethod
    def set(data: Dict, path: str, value: any):
        parts = path.split('.')
        for part in parts[:-1]:
            if part not in data:
                data[part] = {}
            data = data[part]
        data[parts[-1]] = value

    @staticmethod
    def delete(data: Dict, path: str):
        parts = path.split('.')
        for part in parts[:-1]:
            if part not in data:
                return
            data = data[part]
        key = parts[-1]
        if key not in data:
            return
        del data[parts[-1]]



from __future__ import annotations

import argparse

from octoploy.backup.BackupGenerator import BackupGenerator
from octoploy.config.Config import ProjectConfig, RunMode
from octoploy.deploy.AppDeploy import AppDeployment
from octoploy.utils.Log import Log

log_instance = Log('octoploy')


def load_project(config_dir: str) -> ProjectConfig:
    if config_dir != '':
        return ProjectConfig.load(config_dir)

    # No path specified, try a few common ones
    paths = ['.', 'configs', 'octoploy']
    for path in paths:
        try:
            return ProjectConfig.load(path)
        except FileNotFoundError:
            continue
    raise FileNotFoundError(f'Did not find config in any of {paths}')


def reload_config(args):
    root_config = load_project(args.config_dir)
    app_config = root_config.load_app_config(args.name[0])
    if not app_config.enabled():
        log_instance.log.error('App is disabled')
        return
    if app_config.is_template():
        log_instance.log.error('App is a template')
        return

    oc = root_config.create_oc()
    log_instance.log.info('Reloading ' + app_config.get_dc_name())
    reload_actions = app_config.get_reload_actions()
    for action in reload_actions:
        action.run(oc)
    log_instance.log.info('Done')


def _run_app_deploy(config_dir: str, app_name: str, mode: RunMode):
    root_config = load_project(config_dir)
    app_config = root_config.load_app_config(app_name)
    deployment = AppDeployment(root_config, app_config, mode)
    deployment.deploy()
    log_instance.log.info('Done')


def _run_apps_deploy(config_dir: str, mode: RunMode):
    root_config = load_project(config_dir)
    configs = root_config.load_app_configs()
    log_instance.log.info(f'Got {len(configs)} configs')
    for app_config in configs:
        AppDeployment(root_config, app_config, mode).deploy()
    log_instance.log.info('Done')


def plan_app(args):
    mode = RunMode()
    mode.plan = True
    _run_app_deploy(args.config_dir, args.name[0], mode)


def deploy_app(args):
    mode = RunMode()
    mode.out_file = args.out_file
    mode.dry_run = args.dry_run
    _run_app_deploy(args.config_dir, args.name[0], mode)


def plan_all(args):
    mode = RunMode()
    mode.plan = True
    _run_apps_deploy(args.config_dir, mode)


def deploy_all(args):
    mode = RunMode()
    mode.out_file = args.out_file
    mode.dry_run = args.dry_run
    _run_apps_deploy(args.config_dir, mode)


def create_backup(args):
    root_config = load_project(args.config_dir)
    BackupGenerator(root_config).create_backup(args.name[0])


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', dest='debug', action='store_true')
    parser.add_argument('-c', '--config-dir', dest='config_dir',
                        help='Path to the folder containing all configurations',
                        default='')

    subparsers = parser.add_subparsers(help='Commands')
    backup_parser = subparsers.add_parser('backup', help='Creates a backup of all resources in the cluster')
    backup_parser.add_argument('name', help='Name of the backup folder', nargs=1)
    backup_parser.set_defaults(func=create_backup)

    reload_parser = subparsers.add_parser('reload', help='Reloads the configuration of a running application')
    reload_parser.add_argument('name', help='Name of the app which should be reloaded (folder name)', nargs=1)
    reload_parser.set_defaults(func=reload_config)

    plan_parser = subparsers.add_parser('plan', help='Verifies what changes have to be applied for a single app')
    plan_parser.add_argument('name', help='Name of the app which should be checked (folder name)', nargs=1)
    plan_parser.set_defaults(func=plan_app)

    plan_all_parser = subparsers.add_parser('plan-all',
                                            help='Verifies what changes have to be applied for all apps')
    plan_all_parser.set_defaults(func=plan_all)

    deploy_parser = subparsers.add_parser('deploy', help='Deploys the configuration of an application')
    deploy_parser.add_argument('--out-file', dest='out_file',
                               help='Writes all objects into a yml file instead of deploying them. '
                                    'This does not communicate with openshift in any way')
    deploy_parser.add_argument('--dry-run', dest='dry_run', help='Does not interact with openshift',
                               action='store_true')
    deploy_parser.add_argument('name', help='Name of the app which should be deployed (folder name)', nargs=1)
    deploy_parser.set_defaults(func=deploy_app)

    deploy_all_parser = subparsers.add_parser('deploy-all',
                                              help='Deploys all objects of all enabled application')
    deploy_all_parser.add_argument('--out-file', dest='out_file',
                                   help='Writes all objects into a yml file instead of deploying them. '
                                        'This does not communicate with openshift in any way')
    deploy_all_parser.add_argument('--dry-run', dest='dry_run', help='Does not interact with openshift',
                                   action='store_true')
    deploy_all_parser.set_defaults(func=deploy_all)

    args = parser.parse_args()
    if 'func' not in args.__dict__:
        parser.print_help()
        exit(1)
        return

    if args.debug:
        Log.set_debug()

    args.func(args)


if __name__ == '__main__':
    main()



from unittest import TestCase

import yaml

from octoploy.processing.OcObjectMerge import OcObjectMerge


class OcObjectMergeTest(TestCase):

    def test_append_sidecar_container(self):
        existing = '''kind: DeploymentConfig
apiVersion: v1
metadata:
  name: "${DC_NAME}"

spec:
  replicas: 1
  selector:
    name: "${DC_NAME}"
  strategy:
     type: Rolling

  template:
    metadata:
      labels:
        name: "${DC_NAME}"

    spec:
      containers:
        - name: "${DC_NAME}"
          volumeMounts:
            - name: java-cacerts-volume
              mountPath: /java/java-cacerts
'''
        new = '''
kind: DeploymentConfig
spec:
  template:
    spec:
      containers:
        - name: "sidecar-container"
'''
        merge = OcObjectMerge()
        data = yaml.safe_load(existing)
        merge.merge(data, yaml.safe_load(new))

        def validate(data):
            self.assertEqual('v1', data['apiVersion'])
            containers = data['spec']['template']['spec']['containers']
            self.assertEqual(2, len(containers))

            for container in containers:
                if container['name'] == '${DC_NAME}':
                    self.assertEqual(1, len(container['volumeMounts']))
                else:
                    self.assertEqual('sidecar-container', container['name'])

        validate(data)

        # Now invert the inputs, the result should be the same
        new_data = yaml.safe_load(new)
        merge.merge(new_data, yaml.safe_load(existing))
        validate(new_data)

    def test_merge_same_container(self):
        existing = '''kind: DeploymentConfig
apiVersion: v1
metadata:
  name: "${DC_NAME}"

spec:
  replicas: 1
  selector:
    name: "${DC_NAME}"
  strategy:
     type: Rolling

  template:
    metadata:
      labels:
        name: "${DC_NAME}"

    spec:
      containers:
        - name: "${DC_NAME}"
          imagePullPolicy: Always
          resources:
            requests:
              cpu: "20m"
              memory: "10Mi"
            limits:
              memory: "300Mi"
          envFrom: 
            - configMapRef:
                name: ms-config
            - secretRef:
                name: ms-secrets

          volumeMounts:
            - name: java-cacerts-volume
              mountPath: /java/java-cacerts

      volumes:
        - name: java-cacerts-volume
          configMap:
            name: java-cacerts

  triggers:
    - type: ConfigChange
    - type: ImageChange
      imageChangeParams:
        automatic: true
        containerNames:
          - "${DC_NAME}"
        from:
          kind: ImageStreamTag
          name: '${DC_NAME}:prod'
'''
        new = '''
kind: DeploymentConfig
metadata:
  name: "${DC_NAME}"

spec:
  template:
    metadata:
      labels:
        name: "${DC_NAME}"

    spec:
      containers:
        - name: "${DC_NAME}"
          volumeMounts:
            - name: crawler-packages-volume
              mountPath: /packages

      volumes:
        - name: crawler-packages-volume
          persistentVolumeClaim:
            claimName: "pvc-crawler-packages"'''

        merge = OcObjectMerge()
        data = yaml.safe_load(existing)
        merge.merge(data, yaml.safe_load(new))

        containers = data['spec']['template']['spec']['containers']
        self.assertEqual(1, len(containers))

        container = containers[0]
        self.assertEqual(2, len(container['volumeMounts']))

        volumes = data['spec']['template']['spec']['volumes']
        self.assertEqual(2, len(volumes))



from typing import Optional

import yaml


class YmlConfig:
    def __init__(self, path: Optional[str]):
        self.data = {}
        self._path = path
        if path is not None:
            with open(path, 'r') as stream:
                self.data = yaml.safe_load(stream)



import json
import platform
import subprocess
from abc import abstractmethod
from typing import Optional, List

from octoploy.oc.Model import ItemDescription, PodData
from octoploy.utils.Log import Log


class K8Api(Log):
    def __init__(self):
        super().__init__('K8Api')

    @abstractmethod
    def tag(self, source: str, dest: str):
        """
        Tags the given image stream (OC only!)
        :param source: Source tag
        :param dest: Destination tag
        """
        raise NotImplemented

    def get_namespaces(self) -> List[str]:
        """
        Returns all namespaces
        :return: Namespaces
        """

    @abstractmethod
    def get(self, name: str) -> Optional[ItemDescription]:
        """
        Returns the given item
        :param name: Name
        :return: Data (if found)
        """
        raise NotImplemented

    @abstractmethod
    def apply(self, yml: str) -> str:
        """
        Applies the given yml file
        :param yml: Yml file
        :return: Stdout
        """
        raise NotImplemented

    @abstractmethod
    def get_pod(self, dc_name: str = None, pod_name: str = None) -> Optional[PodData]:
        """
        Returns a pod by deployment name or pod name
        :param dc_name: Deployment name
        :param pod_name: Pod name
        :return: Pod (if found)
        :raise Exception: More than one pod found
        """
        raise NotImplemented

    @abstractmethod
    def get_pods(self, dc_name: str = None, pod_name: str = None) -> List[PodData]:
        """
        Returns all pods which match the given deployment name or pod name
        :param dc_name: Deployment name
        :param pod_name: Pod name
        :return: Pods
        """
        raise NotImplemented

    @abstractmethod
    def rollout(self, name: str):
        """
        Re-Deploys the latest DC with the given name
        :param name: Deployment name
        """
        raise NotImplemented

    @abstractmethod
    def exec(self, pod_name: str, cmd: str, args: List[str]):
        """
        Executes a command in the given pod
        :param pod_name: Pod name
        :param cmd: Command
        :param args: Arguments
        """
        raise NotImplemented

    @abstractmethod
    def project(self, project: str):
        """
        Changes the default project / namespace
        :param project: Project
        """
        raise NotImplemented

    @abstractmethod
    def switch_context(self, context: str):
        """
        Changes the configuration context
        :param context: Context
        """
        raise NotImplemented

    @abstractmethod
    def annotate(self, name: str, key: str, value: str):
        """
        Add / updates the annotation at the given item
        :param name: Name
        :param key: Annotation key
        :param value: Annotation value
        """
        raise NotImplemented


class Oc(K8Api):
    def get_namespaces(self) -> List[str]:
        lines = self._exec(['get', 'namespaces', '-o', 'name'])
        return lines.splitlines()

    def tag(self, source: str, dest: str):
        self._exec(['tag', source, dest], print_out=True)

    def get(self, name: str) -> Optional[ItemDescription]:
        try:
            json_str = self._exec(['get', name, '-o', 'json'])
        except Exception as e:
            if 'NotFound' in str(e):
                return None
            raise

        return ItemDescription(json.loads(json_str))

    def apply(self, yml: str) -> str:
        return self._exec(['apply', '-f', '-'], stdin=yml)

    def get_pod(self, dc_name: str = None, pod_name: str = None) -> Optional[PodData]:
        pods = self.get_pods(dc_name=dc_name, pod_name=pod_name)
        if len(pods) == 0:
            return None
        if len(pods) > 1:
            raise Exception('More than one match found')
        return pods[0]

    def get_pods(self, dc_name: str = None, pod_name: str = None) -> List[PodData]:
        pods = []
        json_str = self._exec(['get', 'pods', '-o', 'json'])
        data = json.loads(json_str)
        items = data['items']
        for pod in items:
            metadata = pod['metadata']
            version = int(metadata['annotations'].get('openshift.io/deployment-config.latest-version', 0))
            labels = metadata.get('labels', {})
            name = metadata['name']
            status = pod['status'].get('containerStatuses', [{}])
            if len(status) > 0:
                status = status[0]

            pod_data = PodData()
            pod_data.name = name
            pod_data.version = version
            pod_data.ready = status.get('ready', False)
            pod_data.set_labels(labels)

            if pod_name is not None and pod_data.name != pod_name:
                continue
            if dc_name is not None and pod_data.deployment_config != dc_name:
                continue
            pods.append(pod_data)

        return pods

    def rollout(self, name: str):
        """
        Re-Deploys the latest DC with the given name
        :param name: DC name
        """
        self._exec(['rollout', 'latest', name])

    def exec(self, pod_name: str, cmd: str, args: List[str]):
        proc_args = ['exec', pod_name, '--', cmd]
        proc_args.extend(args)
        self._exec(proc_args, print_out=True)

    def project(self, project: str):
        self._exec(['project', project])

    def switch_context(self, context: str):
        raise NotImplemented('Not available for openshift')

    def annotate(self, name: str, key: str, value: str):
        self._exec(['annotate', '--overwrite=true', name, key + '=' + value])

    def _exec(self, args, print_out: bool = False, stdin: str = None) -> str:
        args.insert(0, self._get_bin())
        if print_out:
            print(str(args))

        stdin_bytes = None
        if stdin is not None:
            stdin_bytes = stdin.encode('utf-8')

        self.log.debug('Executing ' + str(args))
        result = subprocess.run(args, capture_output=True, input=stdin_bytes)
        if result.returncode != 0:
            if stdin is not None:
                print(stdin.replace('\\n', '\n'))
            raise Exception('Failed: ' + str(result.stderr.decode('utf-8')))
        output = result.stdout.decode('utf-8')
        if print_out:
            print(output)
        return output

    def _get_bin(self) -> str:
        if platform.system() == 'Windows':
            return 'oc.exe'
        return 'oc'


class K8(Oc):
    _namespace: str = ''

    def rollout(self, name: str):
        self._exec(['rollout', 'restart', 'deployments', name])

    def tag(self, source: str, dest: str):
        raise NotImplemented('Not available for k8')

    def project(self, project: str):
        self._namespace = project

    def switch_context(self, context: str):
        self._exec(['config', 'use-context', context])

    def _exec(self, args, print_out: bool = False, stdin: str = None):
        if self._namespace != '':
            args.append('--namespace=' + self._namespace)
        return super()._exec(args, print_out, stdin)

    def _get_bin(self) -> str:
        if platform.system() == 'Windows':
            return 'kubectl.exe'
        return 'kubectl'



from unittest import TestCase

from octoploy.utils.DictUtils import DictUtils


class DictUtilsTest(TestCase):
    def test_get(self):
        out = DictUtils.get({
            'a': {
                'b': {
                    'c': 1
                }
            }
        }, 'a.b.c')
        self.assertEqual(1, out)

    def test_get_missing(self):
        out = DictUtils.get({
            'a': {
                'b': {
                    'c': 1
                }
            }
        }, 'a.f')
        self.assertIsNone(out)

    def test_get_none(self):
        out = DictUtils.get(None, 'a.f')
        self.assertIsNone(out)

    def test_set(self):
        obj = {}
        DictUtils.set(obj, 'a.b.c', 1)
        self.assertEqual(1, DictUtils.get(obj, 'a.b.c'))

        obj = {}
        DictUtils.set(obj, 'a', 2)
        self.assertEqual(2, DictUtils.get(obj, 'a'))



import os
from typing import Dict


class ConfigMapObject:
    data: Dict
    disable_templating: bool = False

    def __init__(self, data: Dict, disable_templating: bool = False):
        self.data = data
        self.disable_templating = disable_templating


class ConfigMap:
    def __init__(self, data):
        self._data = data
        self.name = data['name']
        self.files = data['files']
        self.disable_templating = data.get('disableTemplating', False)

    def build_object(self, config_root: str) -> ConfigMapObject:
        """
        Creates an configmap object out of this definition
        """
        config_data = {}
        data = {
            'kind': 'ConfigMap',
            'apiVersion': 'v1',
            'metadata': {
                'name': self.name
            },
            'data': config_data
        }

        for file_obj in self.files:
            file = file_obj['file']
            name = file_obj.get('name', os.path.basename(file))
            with open(os.path.join(config_root, file), 'r') as f:
                content = f.read()
            config_data[name] = content

        return ConfigMapObject(data, self.disable_templating)



from __future__ import annotations

import base64
import os
from abc import abstractmethod
from typing import Dict
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from octoploy.config.BaseConfig import BaseConfig
from octoploy.utils.Cert import Cert
from octoploy.utils.Errors import ConfigError


class ValueLoader:
    def __init__(self, config: BaseConfig):
        self._config = config

    @abstractmethod
    def load(self, data: Dict) -> Dict[str, str]:
        pass


class EnvLoader(ValueLoader):

    def load(self, data: Dict) -> Dict[str, str]:
        return dict(os.environ)


class FileLoader(ValueLoader):

    def load(self, data: Dict) -> Dict[str, str]:
        file = data['file']
        encoding = data.get('encoding', 'utf-8')
        conversion = data.get('conversion')

        with open(file, 'rb') as f:
            content = f.read()

        if conversion is not None:
            if conversion == 'base64':
                return {'': base64.b64encode(content).decode('utf-8')}
            raise ValueError(f'Unknown conversion {conversion}')

        return {'': content.decode(encoding)}


class PemLoader(ValueLoader):

    def load(self, data: Dict) -> Dict[str, str]:
        file = data['file']
        cert = Cert(self._config.get_file(file))
        return {
            '_PUBLIC': cert.cert,
            '_KEY': cert.key,
            '_CACERT': ''.join(cert.cacerts),
        }


class ValueLoaderFactory:

    @staticmethod
    def create(config: BaseConfig, loader_name: str) -> ValueLoader:
        if loader_name == 'pem':
            return PemLoader(config)
        if loader_name == 'file':
            return FileLoader(config)
        if loader_name == 'env':
            return EnvLoader(config)
        raise ConfigError('Unknown loader ' + loader_name)



import os

from octoploy.config.Config import ProjectConfig
from octoploy.utils.Log import Log


class BackupGenerator(Log):
    """
    Very crude backup implementation.
    """

    def __init__(self, config: ProjectConfig):
        super().__init__()
        self._config = config

    def create_backup(self, dir_name: str):
        if not os.path.exists(dir_name):
            os.mkdir(dir_name)

        oc = self._config.create_oc()
        namespaces = oc.get_namespaces()
        apis = oc._exec(['api-resources', '--namespaced', '-o', 'name']).splitlines()
        for namespace in namespaces:
            namespace = namespace.split('/')[1]
            self.log.info(f'Backup up namespace {namespace}')
            oc.project(namespace)
            for api in apis:
                try:
                    names = oc._exec(['get', api, '-o', 'name']).splitlines()
                except Exception:
                    continue
                self.log.info(f'Backing up api {api}')
                for item in names:
                    output = oc._exec(['get', item, '-o', 'yaml'])
                    file_name = namespace + '_' + item.replace('/', '_') + '.yaml'
                    with open(os.path.join(dir_name, file_name), 'w') as f:
                        f.write(output)



from __future__ import annotations

import os
from typing import List

import yaml

from octoploy.config.Config import ProjectConfig, AppConfig, RunMode
from octoploy.deploy.DeploymentBundle import DeploymentBundle
from octoploy.deploy.OcObjectDeployer import OcObjectDeployer
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor
from octoploy.utils.Log import Log


class AppDeployment:
    """
    Deploys a single application
    """

    def __init__(self, root_config: ProjectConfig, app_config: AppConfig, mode: RunMode):
        self._root_config = root_config
        self._app_config = app_config
        self._mode = mode

    def deploy(self):
        """
        Deploys all instances of the app
        """
        if self._mode.out_file is not None:
            if os.path.isfile(self._mode.out_file):
                os.remove(self._mode.out_file)

        factory = AppDeployRunnerFactory(self._root_config, self._mode)
        runners = factory.create(self._app_config)
        for runner in runners:
            runner.deploy()


class AppDeployRunnerFactory:
    """
    Creates AppDeployRunner objects
    """

    def __init__(self, root_config: ProjectConfig, mode: RunMode):
        self._root_config = root_config
        self._mode = mode

    def create(self, root_app_config: AppConfig) -> List[AppDeployRunner]:
        """
        Creates deployment runner instances for the given app
        :param root_app_config: App for which the instances should be created
        """
        runners = []
        for app_config in root_app_config.get_for_each():
            runner = AppDeployRunner(self._root_config, app_config, mode=self._mode)
            runners.append(runner)
        return runners


class AppDeployRunner(Log):
    """
    Executes the deployment of a single app
    """

    def __init__(self, root_config: ProjectConfig, app_config: AppConfig, mode: RunMode = RunMode()):
        super().__init__()
        self._root_config = root_config
        self._app_config = app_config
        self._bundle = DeploymentBundle(self._root_config.get_pre_processor())
        self._mode = mode

    def deploy(self):
        """
        Deploys all items for the given app
        """
        if not self._app_config.enabled():
            raise ValueError('App is disabled')
        if self._app_config.is_template():
            raise ValueError('App is a template and can\'t be deployed')

        template_processor = self._app_config.get_template_processor()
        template_processor.parent(self._root_config.get_template_processor())

        self._deploy_templates(self._app_config.get_pre_template_refs(), template_processor)
        self._load_files(self._app_config.get_config_root(), template_processor)
        self._deploy_extra_configmaps(template_processor)
        self._deploy_templates(self._app_config.get_post_template_refs(), template_processor)

        k8api = self._root_config.create_oc()
        if self._mode.out_file is not None:
            self._bundle.dump_objects(self._mode.out_file)
        if self._mode.dry_run:
            return

        self.log.info('Checking ' + self._app_config.get_dc_name())
        object_deployer = OcObjectDeployer(self._root_config, k8api, self._app_config, mode=self._mode)
        self._bundle.deploy(object_deployer)

    def _deploy_templates(self, template_names: List[str], template_processor: YmlTemplateProcessor):
        """
        Deploys all referenced templates (recursively)
        """
        for template_name in template_names:
            template = self._root_config.load_app_config(template_name)
            if not template.is_template():
                raise ValueError('Referenced app ' + template_name + ' is not declared as template')
            if not template.enabled():
                self.log.warning('Template ' + template_name + ' is disabled, skipping')
                return

            child_template_processor = YmlTemplateProcessor(template)
            # Inherit all vars from the previous template processor
            # The child processor is a parent from a config perspective
            # since its configuration will be overwritten by the previous template
            child_template_processor.child(template_processor)

            # The template might reference other templates
            # -> Recursively deploy them
            self._deploy_templates(template.get_pre_template_refs(), child_template_processor)
            self._load_files(template.get_config_root(), child_template_processor)
            self._deploy_templates(template.get_post_template_refs(), child_template_processor)

    def _load_files(self, root: str, template_processor: YmlTemplateProcessor):
        """
        Loads all yml files inside the given folder
        :param root: Path to the root of the configs folder
        """
        for item in os.listdir(root):
            path = os.path.join(root, item)
            if not os.path.isfile(path) or not item.endswith('.yml') or item.startswith('_'):
                continue

            with open(path, 'r') as stream:
                data = yaml.load_all(stream, Loader=yaml.FullLoader)
                for doc in data:
                    if doc is None:
                        # Empty block
                        continue
                    self._bundle.add_object(doc, template_processor)

    def _deploy_extra_configmaps(self, template_processor: YmlTemplateProcessor):
        """
        Deploys all defined file based configmaps
        :param template_processor: Template processor which should be applied
        """
        for config in self._app_config.get_config_maps():
            cm_object = config.build_object(self._app_config.get_config_root())
            self._bundle.add_object(cm_object.data, None if cm_object.disable_templating else template_processor)



from __future__ import annotations

import os

import yaml

from octoploy.deploy.OcObjectDeployer import OcObjectDeployer
from octoploy.processing.DataPreProcessor import DataPreProcessor
from octoploy.processing.OcObjectMerge import OcObjectMerge
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor
from octoploy.utils.Log import Log
from octoploy.utils.YmlWriter import YmlWriter


class DeploymentBundle(Log):
    """
    Holds all objects of a single deployment
    """

    def __init__(self, pre_processor: DataPreProcessor):
        super().__init__()
        self.objects = []  # All objects which should be deployed
        self._pre_processor = pre_processor

    def add_object(self, data: dict, template_processor: YmlTemplateProcessor):
        """
        Adds a new object which should be deployed
        :param data: Object
        :param template_processor: Template processor which should be used
        """
        item_kind = data.get('kind', '').lower()
        if item_kind == '':
            self.log.info('Unknown object kind: ' + str(data))
            return
        if item_kind == 'Secret'.lower():
            self.log.info('Secrets are ignored')
            return
        
        # Pre-process any variables
        if template_processor is not None:
            template_processor.process(data)

        merger = OcObjectMerge()
        # Check if the new data can be merged into any existing objects
        for item in self.objects:
            if merger.merge(item, data):
                # Data has been merged
                return

        self._pre_processor.process(data)
        self.objects.append(data)

    def deploy(self, deploy_runner: OcObjectDeployer):
        """
        Deploys all object
        :param deploy_runner: Deployment runner which should be used
        """
        deploy_runner.select_project()

        # First sort the objects
        # we want deploymentconfigs to be the last items since a config change might
        # have an impact
        def sorting(x):
            object_kind = x['kind'].lower()
            if object_kind == 'DeploymentConfig'.lower() or \
                    object_kind == 'Deployment'.lower():
                return 1
            return 0

        self.objects.sort(key=sorting)
        for item in self.objects:
            deploy_runner.deploy_object(item)

    def dump_objects(self, path: str):
        """
        Very crude method for dumping the objects to a yml file.
        If the file does already exist the content will be appended
        :param path: Path to a file
        """
        all_objects = []
        if os.path.isfile(path):
            with open(path) as f:
                data = yaml.load_all(f, Loader=yaml.FullLoader)
                for doc in data:
                    all_objects.append(doc)

        all_objects.extend(self.objects)
        with open(path, 'w') as file:
            YmlWriter.dump_all(all_objects, file)



from typing import Optional, List


class Cert:
    """
    Parses pem files and extracts the cert, cacert, and key
    """

    def __init__(self, pem: str):
        self.key = None  # type: Optional[str]
        self.cert = None  # type: Optional[str]
        self.cacerts = []  # type: List[str]

        with open(pem) as f:
            lines = f.readlines()

        line_buffer = []
        mode = 0
        for line in lines:
            if '-BEGIN CERTIFICATE-' in line:
                line_buffer = [line]
                mode = 1
                continue
            if '-BEGIN PRIVATE KEY-' in line:
                line_buffer = [line]
                mode = 2
                continue
            if mode == 0:
                # Ignore
                continue

            line_buffer.append(line)
            if mode == 2 and '-END PRIVATE KEY-' in line:
                self.key = ''.join(line_buffer)
                mode = 0
                continue

            if mode == 1 and '-END CERTIFICATE-' in line:
                # End of segment
                if self.cert is None:
                    self.cert = ''.join(line_buffer)
                    continue
                self.cacerts.append(''.join(line_buffer))
                mode = 0



import os
from typing import Dict, List, Optional

from octoploy.config.YmlConfig import YmlConfig
from octoploy.processing.ValueLoader import ValueLoaderFactory
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor


class BaseConfig(YmlConfig):
    def __init__(self, path: Optional[str], external_vars: Dict[str, str] = None):
        super().__init__(path)
        self._external_vars = {}  # type: Dict[str, str]
        """
        External variables for the templating engine
        """
        if external_vars is not None:
            self._external_vars = external_vars

        self._replacements = {}  # type: Dict[str, str]
        """
        Caching field
        """

    def get_file(self, path: str) -> str:
        """
        Returns the path to a file inside the dir of the app/project
        :param path: Relative path
        :return: Path
        """
        return os.path.abspath(os.path.join(self._path, os.pardir, path))

    def get_template_processor(self) -> YmlTemplateProcessor:
        return YmlTemplateProcessor(self)

    def get_replacements(self) -> Dict[str, any]:
        """
        Returns all variables which are available for the yml files

        :return: Key, value map
        """
        if len(self._replacements) > 0:
            return self._replacements

        items = self.data.get('vars', {})
        new_items = {}
        for key, value in items.items():
            # Value can be a primitive or object
            if isinstance(value, dict):
                # Is a object, use a loader to load the value
                loader_name = value.get('loader')
                if loader_name is not None:
                    loader = ValueLoaderFactory.create(self, value['loader'])
                    new_values = loader.load(value)
                    for new_key, new_val in new_values.items():
                        new_items[key + new_key] = new_val
                    continue

        items.update(new_items)
        items.update(self._external_vars)
        return items

    def get_params(self) -> List[str]:
        """
        Returns all required parameters

        :return: Names
        """
        return self.data.get('params', [])



from __future__ import annotations

from abc import abstractmethod
from typing import Dict, Optional

from octoploy.utils.DictUtils import DictUtils


class NamedItem:
    def __init__(self, data):
        self.data = data

    @abstractmethod
    def get_name(self) -> str:
        pass


class ItemDescription:
    def __init__(self, data):
        self.data = data

    def get_annotation(self, key: str) -> str:
        item = self.data.get('metadata', {}).get('annotations', {}).get(key)
        return item


class PodData:
    def __init__(self):
        self.name = ''
        self.ready = False
        self.version = 0
        self.deployment_config = None
        """
        Name of the associated deployment config
        """

    def set_labels(self, labels):
        self.deployment_config = labels.get('deploymentconfig')


class DeploymentConfig:
    def __init__(self, data):
        self.data = data

    def get_template(self) -> Optional[dict]:
        """
        Returns the template object
        """
        return DictUtils.get(self.data, 'spec.template')

    def get_template_spec(self) -> Optional[dict]:
        """
        Returns the template spec
        """
        return self.get_template().get('spec')

    def get_template_name(self) -> Optional[str]:
        """
        Returns the name of the template
        :return: Name or None if no name is defined
        """
        return DictUtils.get(self.get_template(), 'metadata.labels.name')

    def get_containers(self) -> Dict[str, DeploymentConfigContainer]:
        items = self.get_template_spec().get('containers', [])
        out = {}
        for data in items:
            volume = DeploymentConfigContainer(data)
            out[volume.get_name()] = volume
        return out

    def get_volumes(self) -> Dict[str, DeploymentConfigVolume]:
        items = self.get_template_spec().get('volumes', [])
        out = {}
        for data in items:
            volume = DeploymentConfigVolume(data)
            out[volume.get_name()] = volume
        return out

    def add_container(self, item: DeploymentConfigContainer):
        self._add_to_list(self.get_template_spec(), 'containers', item.data)

    def add_volume(self, item: DeploymentConfigVolume):
        self._add_to_list(self.get_template_spec(), 'volumes', item.data)

    @staticmethod
    def _add_to_list(data, key: str, new_item: any):
        if key in data:
            data[key].append(new_item)
            return
        data[key] = [new_item]


class DeploymentConfigVolume(NamedItem):
    def get_name(self) -> str:
        return self.data['name']


class DeploymentConfigContainer(NamedItem):
    def get_name(self) -> str:
        return self.data['name']



from __future__ import annotations

from typing import TYPE_CHECKING

from octoploy.utils.Log import Log

if TYPE_CHECKING:
    from octoploy.config.Config import AppConfig

from octoploy.oc.Oc import K8Api


class DeploymentActionConfig(Log):
    """
    Configuration for a single user configurable deployment step
    """

    def __init__(self, app_config: AppConfig, data):
        self._data = data
        self._app_config = app_config

    def run(self, oc: K8Api):
        if self._data == 'deploy':
            oc.rollout(self._app_config.get_dc_name())
            return

        exec_config = self._data.get('exec', None)
        if exec_config is not None:
            cmd = exec_config['command']
            args = exec_config['args']

            dc_name = self._app_config.get_dc_name()
            log.info('Reloading via exec in pods of ' + dc_name)
            pods = oc.get_pods(dc_name=dc_name)
            for pod in pods:
                oc.exec(pod.name, cmd, args)
            return



from __future__ import annotations

from typing import List, Optional, Dict

from octoploy.config.BaseConfig import BaseConfig
from octoploy.config.ConfigMap import ConfigMap
from octoploy.config.DeploymentActionConfig import DeploymentActionConfig
from octoploy.utils.DictUtils import DictUtils
from octoploy.utils.Errors import MissingVar


class AppConfig(BaseConfig):
    """
    Contains the configuration for the deployment of a single app
    """

    def __init__(self, config_root: str, path: Optional[str], external_vars: Dict[str, str] = None):
        super().__init__(path, external_vars)
        self._config_root = config_root

    def get_config_maps(self) -> List[ConfigMap]:
        """
        Returns additional config maps which should contain the content of a file
        """
        return [ConfigMap(data) for data in self.data.get('configmaps', [])]

    def enabled(self) -> bool:
        """
        True if this app is enabled
        """
        return self.data.get('enabled', False)

    def is_template(self) -> bool:
        """
        Indicates if this app is a template
        """
        return self.data.get('type', 'app') == 'template'

    def get_config_root(self) -> str:
        """
        Returns the path to the app config folder
        :return: Folder path
        """
        return self._config_root

    def get_for_each(self) -> List[AppConfig]:
        """
        Returns all instances of this app which should be created.
        :return: Instances, 1 by default
        :raise MissingVar: Gets raised if the data inside forEach is not complete
        """
        instances = []
        for instance_vars in self.data.get('forEach', []):
            assert isinstance(instance_vars, dict)
            dc_name = instance_vars.get('DC_NAME')
            if dc_name is None:
                raise MissingVar('DC_NAME not defined in forEach for app ' + self.get_dc_name())

            config = AppConfig(self._config_root, None, instance_vars)
            # Inherit all parameters
            config.data.update(self.data)
            # Update the DC_NAME
            DictUtils.set(config.data, 'dc.name', dc_name)
            instances.append(config)

        if len(instances) == 0:
            # No forEach defined, just create one instance
            instances.append(self)
        return instances

    def get_pre_template_refs(self) -> List[str]:
        """
        Returns the name of the templates that should be applied before processing own objects

        :return: Template names
        """
        return self.data.get('applyTemplates', [])

    def get_post_template_refs(self) -> List[str]:
        """
        Returns the name of the templates that should be applied after processing own objects

        :return: Template names
        """
        return self.data.get('postApplyTemplates', [])

    def get_reload_actions(self) -> List[DeploymentActionConfig]:
        """
        Returns all actions that should be executed after a configuration change of the app
        """
        return [DeploymentActionConfig(self, x) for x in self.data.get('on-config-change', [])]

    def get_dc_name(self) -> Optional[str]:
        """
        Returns the configured name of the deployment config
        :return: Name
        """
        return DictUtils.get(self.data, 'dc.name')

    def get_replacements(self) -> Dict[str, str]:
        """
        Returns all variables which are available for the yml files
        :return: Key, value map
        """
        items = super().get_replacements()
        dc_name = self.get_dc_name()
        if dc_name is not None:
            items.update({
                'DC_NAME': dc_name
            })
        return items



from __future__ import annotations

import re
from typing import Optional, Dict, List, Set
from typing import TYPE_CHECKING

from octoploy.utils.Log import Log

if TYPE_CHECKING:
    from octoploy.config.BaseConfig import BaseConfig
from octoploy.utils.Errors import MissingParam


class YmlTemplateProcessor(Log):
    """
    Processes yml files by replacing any string placeholders.
    """

    VAR_PATTERN = re.compile(r'\${(.+?)}')
    KEY_FIELD_MERGE: str = '_merge'

    def __init__(self, config: BaseConfig):
        super().__init__()
        self._missing_vars = []  # type: List[str]
        """
        List of all variables which have not been replace because
        there was no value defined for them
        """
        self._config = config  # type: BaseConfig
        self._parent = None  # type: Optional[YmlTemplateProcessor]
        self._child = None  # type: Optional[YmlTemplateProcessor]

    def process(self, data: dict):
        """
        Processes the app data

        :param data: Data of the app, the data will be modified in place
        :raise MissingParam: Gets raised if at least one parameter is not defined
        """
        replacements = self._get_replacements()
        # Replace any variables
        depth = 0
        found_var = True
        while depth < 10 and found_var:  # Lazily assume there are only 10 levels of chained reference
            found_var = False
            for key, value in replacements.items():
                if isinstance(value, dict):
                    # A replacement might be an object
                    # So we need to walk though every item and see if there is someting to replace
                    self._walk_dict(replacements, value)
                    continue
                if not isinstance(value, str):
                    continue

                # The replacement value might refer to another variable
                for variable_name in self.VAR_PATTERN.findall(value):
                    new_value = replacements.get(variable_name)
                    if new_value is None:
                        self.log.warning('Missing referenced variable: ' + variable_name)
                        continue
                    if value == '${' + variable_name + '}':
                        # Replace the entire value since the replacement value only consists of the ${} tag
                        replacements[key] = new_value
                        found_var = True
                        continue

                    replacements[key] = value.replace('${' + variable_name + '}', str(new_value))
                    found_var = True

        self._walk_dict(replacements, data)

        # Check if any of the missing vars are declared as "params"
        # (aka are required)
        if len(self._missing_vars) > 0:
            missing_params = []
            params = self._get_params()
            for missing in self._missing_vars:
                if missing not in params:
                    continue
                missing_params.append(missing)
            if len(missing_params) > 0:
                raise MissingParam('The following params are not defined: ' + str(missing_params))
            self.log.warning('The following vars are not defined: ' + str(self._missing_vars))

        self._sanity_check(data)

    def _sanity_check(self, data: Dict[str, any]):
        if '${' in str(data):
            self.log.warning('At least one variable could not been resolved: ' + str(data))

    def _get_params(self) -> Set[str]:
        """
        Returns all defined params
        :return: Param names
        """
        params = set()
        if self._parent is not None:
            params.update(self._parent._get_params())
        params.update(self._config.get_params())
        if self._child is not None:
            params.update(self._child._get_params())
        return params

    def _get_replacements(self) -> Dict[str, any]:
        """
        Returns all replacements handled by this processor, including all parent variables
        :return: Replacements
        """
        replacements = {}
        if self._parent is not None:
            replacements.update(self._parent._get_replacements())
        replacements.update(self._config.get_replacements())
        if self._child is not None:
            replacements.update(self._child._get_replacements())
        return replacements

    def _walk_dict(self, replacements: Dict[str, any], data: dict) -> Dict[str, any]:
        """
        Walks through all items in the dict and replaces any known variables

        :param replacements: Data which should be used as a replacement
        :param data: Data which should be walked
        """
        for key in list(data.keys()):
            obj = data[key]
            if key == self.KEY_FIELD_MERGE:
                del data[key]
                new_dict = self._walk_item(replacements, obj)
                if isinstance(new_dict, str):
                    raise ValueError('No replacement found for ' + new_dict)
                data.update(new_dict)
                continue

            data[key] = self._walk_item(replacements, obj, data, key)
        return data

    def _walk_item(self, replacements: Dict[str, any], obj: any, parent: dict = None, child_key: str = None) -> any:
        if isinstance(obj, list):
            for idx, item in enumerate(obj):
                obj[idx] = self._walk_item(replacements, item)
            return obj

        if isinstance(obj, str):
            return self._replace(obj, replacements)

        if isinstance(obj, dict):
            return self._walk_dict(replacements, obj)
        return obj

    def _replace(self, item: str, replacements: Dict[str, any]) -> any:
        for variable, value in replacements.items():
            if item == '${' + variable + '}':
                # Item only contains a tag, simple replace (non textual)
                return value
            # The variable tag is surrounded by other str or other tags
            item = item.replace('${' + variable + '}', str(value))

        # Search for any missing variables
        if isinstance(item, str):
            self._missing_vars.extend(self.VAR_PATTERN.findall(item))
        return item

    def parent(self, template_processor: YmlTemplateProcessor):
        """
        Inherits all replacements from the given processor.
        If the same value is defined in this and the parent, the definition of the this processor will override
        the parent definition

        :param template_processor: Child processor
        """
        if self._parent is not None:
            raise ValueError('Parent processor already defined')
        self._parent = template_processor

    def child(self, template_processor: YmlTemplateProcessor):
        """
        Inherits all replacements from the given processor.
        If the same value is defined in this and the child, the definition of the child processor will override
        this definition

        :param template_processor: Child processor
        """
        if self._child is not None:
            raise ValueError('Child processor already defined')
        self._child = template_processor



import logging
import sys


class ColorFormatter(logging.Formatter):
    grey = '\x1b[38;21m'
    yellow = '\x1b[33;21m'
    red = '\x1b[31;21m'
    bold_red = '\x1b[31;1m'
    reset = '\x1b[0m'
    format = '%(levelname)-.1s %(asctime)-.19s [%(name)s] %(message)s'

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


class Log:
    log_level = logging.INFO

    def __init__(self, name: str = None):
        if not hasattr(self, 'log'):
            if name is None:
                name = self.__class__.__name__
            log = logging.getLogger(name)
            if not log.hasHandlers():
                handler = logging.StreamHandler(stream=sys.stdout)
                handler.setFormatter(ColorFormatter())
                log.addHandler(handler)
                log.setLevel(Log.log_level)
            self.log = log

    @classmethod
    def set_debug(cls):
        cls.log_level = logging.DEBUG



from unittest import TestCase

from octoploy.config.BaseConfig import BaseConfig
from octoploy.processing.ValueLoader import ValueLoaderFactory


class ValueLoaderTest(TestCase):

    def test_env(self):
        factory = ValueLoaderFactory()
        loader = factory.create(BaseConfig(None), 'env')
        items = loader.load({})
        self.assertTrue(len(items) > 0)

    def test_file(self):
        factory = ValueLoaderFactory()
        loader = factory.create(BaseConfig(None), 'file')
        items = loader.load({'file': __file__})
        self.assertIsNotNone(items.get(''))
        self.assertIn('class', items.get(''))

        items = loader.load({'file': __file__, 'conversion': 'base64'})
        self.assertNotIn('class', items.get(''))



from typing import Dict

from octoploy.oc.Model import DeploymentConfig, NamedItem
from octoploy.utils.DictUtils import DictUtils
from octoploy.utils.Log import Log


class OcObjectMerge(Log):
    """
    Merges openshift objects
    """
    NAME_PATH = 'metadata.name'

    def __init__(self):
        super().__init__()
        self._existing_dc = None  # type: DeploymentConfig
        self._new_dc = None  # type: DeploymentConfig

    def merge(self, existing, to_add) -> bool:
        """
        Merges the given openshift object into one.
        :param existing: Existing where the data should be added to
        :param to_add: New data
        :return: True if the data has been merged, false otherwise
        """
        expected_type = existing['kind'].lower()
        type_str = to_add['kind'].lower()
        if expected_type != type_str:
            return False

        expected_name = DictUtils.get(existing, self.NAME_PATH)
        name = DictUtils.get(to_add, self.NAME_PATH)
        if name is not None and expected_name is not None and expected_name != name:
            return False

        if expected_type == 'DeploymentConfig'.lower():
            self._merge_dc(DeploymentConfig(existing), DeploymentConfig(to_add))
            return True

        self.log.warning('Don\'t know how to merge ' + expected_type)

    def _merge_dc(self, existing: DeploymentConfig, to_add: DeploymentConfig):
        """
        Merges the given deployment configs if their name match
        :param existing: Deployment config where the data should be added
        :param to_add: New data
        """
        self._existing_dc = existing
        self._new_dc = to_add
        to_add_template_name = to_add.get_template_name()
        existing_template_name = existing.get_template_name()
        # If the name of a template is not defined it will be merged regardless of the name of the existing template
        if to_add_template_name is not None and existing_template_name is not None \
                and existing.get_template_name() != to_add_template_name:
            return

        # Containers and volumes need special treatment
        self._merge_object(existing.data, to_add.data, {
            'spec': {'template': {'spec': {
                'containers': True,
                'volumes': True
            }}}})

        self._merge_named_item(existing.get_containers(), to_add.get_containers(), existing.add_container)
        self._merge_named_item(existing.get_volumes(), to_add.get_volumes(), existing.add_volume)

    def _merge_object(self, existing: dict, new_data: dict, ignore_keys: Dict[str, any]):
        """
        Merges the given dict
        :param existing: Dict where the data should be merged to
        :param new_data: New data
        """
        for key, item in new_data.items():
            if ignore_keys.get(key, False) is True:
                continue
            if key not in existing:
                existing[key] = item
                continue
            parent_item = existing[key]
            if isinstance(item, list):
                parent_item.extend(item)
                continue
            if isinstance(item, dict):
                self._merge_object(parent_item, item, ignore_keys.get(key, {}))
                continue
            if item == parent_item:
                continue

            existing[key] = item
            self.log.warning(f'Value conflict: {item} (from {self._new_dc.get_template_name()}) replaces ' + \
                             f'{parent_item} (from {self._existing_dc.get_template_name()})')

    def _merge_named_item(self, existing_items: Dict[str, NamedItem], new_item: Dict[str, NamedItem], add_func_call):
        """
        Merges a named item object
        :param existing_items: Parent items, mapped to their name
        :param new_item: New items, mapped to their name
        :param add_func_call: Function call for adding an item to the parent
        """
        for name, new_container in new_item.items():
            if name in existing_items:
                # Merge data
                self._merge_object(existing_items[name].data, new_container.data, {})
                continue

            # Additional item
            add_func_call(new_container)



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='octoploy',
    version='1.0.7',
    author='davidgiga1993',
    author_email='david@dev-core.org',
    description='Simple kubernetes / openshift templating engine with state tracking, backups and more',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/davidgiga1993/octoploy',
    packages=setuptools.find_packages(),
    install_requires=['pyyaml'],
    entry_points={
        'console_scripts': ['octoploy=octoploy.octoploy:main'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries'
    ],
)



import os
from unittest import TestCase

import yaml

from octoploy.config.Config import ProjectConfig, RunMode
from octoploy.deploy.AppDeploy import AppDeployment
from octoploy.utils.Errors import MissingParam


class AppDeploymentTest(TestCase):

    def setUp(self) -> None:
        self._base_path = os.path.dirname(__file__)
        self._tmp_file = 'out.yml'
        self._mode = RunMode()
        self._mode.dry_run = True
        self._mode.out_file = self._tmp_file

    def tearDown(self) -> None:
        if os.path.isfile(self._tmp_file):
            os.remove(self._tmp_file)

    def test_cm_types(self):
        prj_config = ProjectConfig.load(os.path.join(self._base_path, 'app_deploy_test'))
        app_config = prj_config.load_app_config('cm-types')
        runner = AppDeployment(prj_config, app_config, self._mode)
        runner.deploy()

        with open(self._tmp_file) as f:
            content = f.read()
        # Make sure the "y" is quoted
        self.assertIn('"y"', content)

    def test_library(self):
        prj_config = ProjectConfig.load(os.path.join(self._base_path, 'lib-usage'))
        app_config = prj_config.load_app_config('some-app')
        runner = AppDeployment(prj_config, app_config, self._mode)
        runner.deploy()
        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertEqual('paramValue', data['metadata']['name'])

    def test_var_loader(self):
        prj_config = ProjectConfig.load(os.path.join(self._base_path, 'lib-usage'))
        app_config = prj_config.load_app_config('var-loader-app')
        runner = AppDeployment(prj_config, app_config, self._mode)
        runner.deploy()
        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        expected = '''-----BEGIN PRIVATE KEY-----
KEY STUFF
-----END PRIVATE KEY-----
'''
        self.assertEqual(expected, data['KEY'])

    def test_for_each(self):
        prj_config = ProjectConfig.load(os.path.join(self._base_path, 'app_deploy_test'))
        app_config = prj_config.load_app_config('app-for-each')
        runner = AppDeployment(prj_config, app_config, self._mode)
        runner.deploy()

        docs = []
        with open(self._tmp_file) as f:
            data = yaml.load_all(f, Loader=yaml.FullLoader)
            for doc in data:
                docs.append(doc)

        # We should have two instances
        self.assertEqual(2, len(docs))
        self.assertEqual('hello', docs[0]['metadata']['REMAPPED'])
        self.assertEqual('hello', docs[1]['metadata']['REMAPPED'])

    def test_params(self):
        prj_config = ProjectConfig.load(os.path.join(self._base_path, 'app_deploy_test'))
        app_config = prj_config.load_app_config('app-params')
        runner = AppDeployment(prj_config, app_config, self._mode)
        try:
            runner.deploy()
            self.fail('No exception raised for missing param')
        except MissingParam:
            pass

        external_params = {
            'SomeParam': 'input'
        }
        app_config._external_vars = external_params
        runner.deploy()

        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertEqual('input', data['someObj']['param'])

    def test_inherit_vars(self):
        prj_config = ProjectConfig.load(os.path.join(self._base_path, 'app_deploy_test'))
        app_config = prj_config.load_app_config('app')
        runner = AppDeployment(prj_config, app_config, self._mode)
        runner.deploy()

        with open(self._tmp_file) as f:
            data = yaml.load(f, Loader=yaml.FullLoader)
        self.assertEqual('ABC', data['metadata']['name'])
        self.assertEqual('ABC-1', data['metadata']['REMAPPED'])
        self.assertEqual('ABC', data['metadata']['REMAPPED2'])
        self.assertEqual('DEF', data['metadata']['name2'])
        self.assertEqual('3', data['metadata']['base'])
        self.assertEqual('global', data['metadata']['GLOBAL_TEST'])



from octoploy.utils.DictUtils import DictUtils


class DataPreProcessor:
    def __init__(self):
        pass

    def process(self, yml):
        """
        Preprocesses the yml in place
        """


class OcToK8PreProcessor(DataPreProcessor):

    def process(self, yml):
        kind = yml.get('kind')
        if kind == 'DeploymentConfig':
            yml['kind'] = 'Deployment'
            version = yml.get('apiVersion')
            if not version.startswith('apps/'):
                yml['apiVersion'] = 'apps/' + version

            name_selector = DictUtils.get(yml, 'spec.selector.name')
            if name_selector is not None:
                DictUtils.delete(yml, 'spec.selector.name')
                DictUtils.set(yml, 'spec.selector.matchLabels.app', name_selector)

            strategy_type = DictUtils.get(yml, 'spec.strategy.type')
            if strategy_type == 'Rolling':
                DictUtils.set(yml, 'spec.strategy.type', 'RollingUpdate')

            name = DictUtils.get(yml, 'spec.template.metadata.labels.name')
            if name is not None:
                DictUtils.delete(yml, 'spec.template.metadata.labels.name')
                DictUtils.set(yml, 'spec.template.metadata.labels.app', name)

            return



from unittest import TestCase, mock

from octoploy.config.Config import AppConfig
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor


class YmlTemplateProcessorTest(TestCase):

    def test_replace(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
dc:
    name: hello
vars:
    MY_VAR: testVal
    IMAGE_NAME: image
    MY_OBJECT:
        someItem: 1
        someOtherItem: 2
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {'root': {
            'item': '${DC_NAME}',
            'object': '${MY_OBJECT}',
            'list': [{
                'other': '${DC_NAME}/${IMAGE_NAME}'
            }],
            'sub': {
                'item2': '${MY_VAR}'
            }
        }}
        proc.process(data)
        self.assertEqual('hello', data['root']['item'])
        self.assertEqual('hello/image', data['root']['list'][0]['other'])
        self.assertEqual('testVal', data['root']['sub']['item2'])
        self.assertEqual({
            'someItem': 1,
            'someOtherItem': 2
        }, data['root']['object'])

    def test_params(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
dc:
    name: hello
vars:
    MY_VAR: testVal
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {'root': {
            'item': '${DC_NAME}',
            'list': [{
                'other': '${DC_NAME}'
            }],
            'sub': {
                'item2': '${MY_VAR}'
            }
        }}
        proc.process(data)
        self.assertEqual('hello', data['root']['item'])
        self.assertEqual('hello', data['root']['list'][0]['other'])
        self.assertEqual('testVal', data['root']['sub']['item2'])

    def test_merge_inline(self):
        with mock.patch('builtins.open', mock.mock_open(read_data='''
dc:
    name: hello
vars:
    MERGE_OBJ: 
        someKey: value
''')):
            app_config = AppConfig('', '')

        proc = YmlTemplateProcessor(app_config)
        data = {'root': {
            'item': '${DC_NAME}',
            '_merge': '${MERGE_OBJ}',
        }}
        proc.process(data)
        self.assertEqual('hello', data['root']['item'])
        self.assertEqual('value', data['root']['someKey'])



from __future__ import annotations

import os
from typing import Optional, Dict, List

from octoploy.config.AppConfig import AppConfig
from octoploy.config.BaseConfig import BaseConfig
from octoploy.oc.Oc import Oc, K8, K8Api
from octoploy.processing.DataPreProcessor import DataPreProcessor, OcToK8PreProcessor
from octoploy.processing.YmlTemplateProcessor import YmlTemplateProcessor
from octoploy.utils.Errors import ConfigError


class RunMode:
    def __init__(self):
        self.out_file = None  # type: Optional[str]
        """
        Yml output file
        """

        self.dry_run = False
        """
        True if no OC should be called
        """

        self.plan = False
        """
        True if changes should be previewed
        """


class ProjectConfig(BaseConfig):
    """
    Project configuration
    """

    def __init__(self, config_root: str, path: str):
        super().__init__(path)
        self._config_root = config_root
        self._oc = None
        self._library = None  # type: Optional[ProjectConfig]

        inherit = self.data.get('inherit')
        if inherit is not None:
            # Use a library
            parent_dir = os.path.abspath(os.path.join(path, os.pardir, os.pardir))
            lib_dir = os.path.join(parent_dir, inherit)
            if not os.path.isdir(lib_dir):
                raise FileNotFoundError('Library not found: ' + lib_dir)
            self._library = ProjectConfig.load(lib_dir)
            if not self._library.is_library():
                raise ConfigError('Project ' + inherit + ' referenced as library but is not a library')

    @classmethod
    def load(cls, path: str) -> ProjectConfig:
        return ProjectConfig(path, os.path.join(path, '_root.yml'))

    def get_config_root(self) -> str:
        return self._config_root

    def is_library(self) -> bool:
        """
        Indicates if this collection is a library
        """
        return self.data.get('type', '') == 'library'

    def get_pre_processor(self) -> DataPreProcessor:
        """
        Returns the pre processor for the current config
        """
        mode = self._get_mode()
        if mode == 'k8':
            return OcToK8PreProcessor()
        return DataPreProcessor()

    def _get_mode(self) -> str:
        return self.data.get('mode', 'oc')

    def create_oc(self) -> K8Api:
        """
        Creates a new openshift / k8 client
        :return: Client
        """
        if self._oc is not None:
            return self._oc

        mode = self._get_mode()
        if mode == 'oc':
            oc = Oc()
        elif mode == 'k8':
            oc = K8()
        else:
            raise ValueError(f'Invalid mode: {mode}')
        self._oc = oc
        return oc

    def get_oc_project_name(self) -> Optional[str]:
        """
        Returns the name of the openshift project

        :return: Name or null for libraries
        """
        return self.data.get('project')

    def get_oc_context(self) -> Optional[str]:
        """
        Returns the configuration context name
        :return: Name
        """
        return self.data.get('context')

    def get_template_processor(self) -> YmlTemplateProcessor:
        root_processor = super().get_template_processor()
        if self._library is not None:
            processor = self._library.get_template_processor()
            root_processor.parent(processor)
        return root_processor

    def get_replacements(self) -> Dict[str, str]:
        """
        Returns all variables which are available for the yml files
        :return: Key, value map
        """
        items = super().get_replacements()
        name = self.get_oc_project_name()
        if name is not None:
            items.update({
                'OC_PROJECT': name
            })
        return items

    def load_app_configs(self) -> List[AppConfig]:
        """
        Loads all app configurations available in this project
        :return:
        """
        items = []
        for dir_item in os.listdir(self._config_root):
            path = os.path.join(self._config_root, dir_item)
            if not os.path.isdir(path):
                continue
            try:
                app_config = self.load_app_config(dir_item)
            except FileNotFoundError:
                # Index file missing
                continue

            if app_config.is_template() or not app_config.enabled():
                # Silently skip
                continue
            items.append(app_config)
        if self._library is not None:
            items.extend(self._library.load_app_configs())
        return items

    def load_app_config(self, name: str) -> AppConfig:
        folder_path = os.path.join(self._config_root, name)
        if not os.path.isdir(folder_path):
            if self._library is not None:
                return self._library.load_app_config(name)
            raise FileNotFoundError('App folder not found: ' + folder_path)

        index_file = os.path.join(folder_path, '_index.yml')
        if not os.path.isfile(index_file):
            # Index file missing
            raise FileNotFoundError('No index yml file found: ' + index_file)

        variables = self.get_replacements()
        return AppConfig(folder_path, index_file, variables)
